# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

__all__ = [
    "ConfigurationBaseException",
    "ConfigKeyDoesNotExist",
    "ConfigKeyAlreadyExist",
    "ConfigValueOutOfRange",
    "ConfigFileTypeNotSpecified",
    "ConfigItem",
    "NumericalRange",
    "Variable",
    "IntVariable",
    "FloatVariable",
    "StrVariable",
    "Configuration",
    "ConfigurationControl",
    "setting"
]

# std
import copy
import json
import inspect
import threading

from base64 import b64encode, b64decode
from typing import Iterable, Mapping, Optional, Type
from dataclasses import dataclass, field

# site
from typex import Static


class ConfigurationBaseException (Exception): ...
class ConfigKeyDoesNotExist (ConfigurationBaseException): ...
class ConfigKeyAlreadyExist (ConfigurationBaseException): ...
class ConfigValueOutOfRange (ConfigurationBaseException): ...
class ConfigFileTypeNotSpecified (ConfigurationBaseException): ...
class ConfigFileTypeNotSupported (ConfigurationBaseException): ...


class ConfigFileType (Static):
    JSON = "json"


@dataclass
class ConfigItem (object):
    type_: Type[int | float | str]
    default: Optional[int | float | str] = field(default=None)
    ranges: Optional[Iterable[int | float | str]] = field(default=None)
    value: Optional[int | float | str] = field(default=None)


@dataclass
class NumericalRange (object):
    min: int | float
    max: int | float
    round: int = field(default=0)

    def __post_init__(self):
        if not isinstance(self.min, (int, float)):
            raise TypeError(f"Expected `min` to be `int` or `float`, but got {type(self.min)}.")

        if not isinstance(self.max, (int, float)):
            raise TypeError(f"Expected `max` to be `int` or `float`, but got {type(self.max)}.")

        if not isinstance(self.round, int):
            raise TypeError(f"Expected `round` to be `int`, but got {type(self.round)}.")

        if self.round < 0:
            raise ValueError(f"Expected `round` to be greater than or equal to 0, but got {self.round}.")


@dataclass
class LastState (object):
    filepath: Optional[str] = field(default=None)
    filetype: Optional[str] = field(default=None)
    encoding: str = field(default="utf-8")
    base64: bool = field(default=False)


class Variable (object):
    def __new__(cls, stem: "ConfigurationControl", key: str, value: int | float | str) -> "Variable":
        if cls is Variable:
            raise TypeError("Cannot create an instance of the abstract class Variable.")

        instance = super(Variable, cls).__new__(cls, value)
        instance.__set_attribute(stem, key)

        return instance

    def __set_attribute(self, stem: "ConfigurationControl", key: str) -> None:
        self.__stem = stem
        self.__key = key

    @property
    def ranges(self) -> Optional[Iterable[int | float | str] | NumericalRange]:
        return self.__stem.get_ranges(self.__key)

    @property
    def value(self) -> int | float | str:
        return self.__stem.get(self.__key)

    @property
    def stem(self) -> "Configuration":
        return self.__stem.surface

    @property
    def name(self) -> str:
        return self.__key

    @property
    def types(self) -> Type[int | float | str]:
        return self.__stem.get_type(self.__key)

    @property
    def default(self) -> Optional[int | float | str]:
        return self.__stem.get_default(self.__key)

    def set(self, value: int | float | str) -> None:
        self.__stem.set(self.__key, value)


class IntVariable (Variable, int): ...


class FloatVariable (Variable, float): ...


class StrVariable (Variable, str): ...


class Configuration (object):
    def __init__(self):
        self.ctrl = ConfigurationControl(self)
        items = [x for x in dir(self) if not x.startswith("_") and x != "ctrl"]
        getattribute = super().__getattribute__
        for name in items:
            item = getattribute(name)
            if not isinstance(item, ConfigItem):
                continue

            self.ctrl.new(name, item.type_, item.default, item.ranges)

    def __setattr__(self, name, value) -> None:
        if name.startswith("_") or name == "ctrl":
            return super().__setattr__(name, value)

        self.ctrl.set(name, value)

    def __setitem__(self, name, value) -> None:
        self.ctrl.set(name, value)

    def __getitem__(self, name) -> "Variable":
        return self.ctrl.get(name)

    def __getattribute__(self, name) -> "Variable":
        if name.startswith("_") or name == "ctrl":
            return super().__getattribute__(name)

        return self.ctrl.get(name)


class ConfigurationControl (object):
    def __init__(self, surface: Configuration):
        self.surface = surface
        self._lock = threading.RLock()
        self.__table: Mapping[str, ConfigItem] = {}
        self.__invalid: Mapping[str, int | float | str] = {}
        self.__last = LastState()

    @property
    def last_filepath(self) -> Optional[str]:
        with self._lock:
            return self.__last.filepath

    @property
    def last_filetype(self) -> Optional[str]:
        with self._lock:
            return self.__last.filetype

    @property
    def last_encoding(self) -> str:
        with self._lock:
            return self.__last.encoding

    @property
    def last_base64(self) -> bool:
        with self._lock:
            return self.__last.base64

    def new(self, key: str, type_: Type[int | float | str], default: Optional[int | float | str] = None,
            ranges: Optional[Iterable[int | float | str] | NumericalRange] = None) -> None:
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        if inspect.isclass(type_) and not issubclass(type_, (int, float, str)):
            raise TypeError(f"Expected `type_` to be `int`, `float` or `str` class, but got {type(type_)}.")

        if not isinstance(default, type_) and default is not None:
            raise TypeError(f"Expected `default` to be `{type_}` or None, but got {type(default)}.")

        if not isinstance(ranges, (Iterable, NumericalRange)) and not ranges is None:
            raise TypeError(f"Expected `ranges` to be Iterable or NumericalRange, but got {type(ranges)}.")

        with self._lock:
            if key in self.__table:
                raise ConfigKeyAlreadyExist(f"The config key `{key}` already exists.")

            default = default if default is not None else type_()
            self.__table[key] = ConfigItem(type_, default, ranges, None)

    def set(self, key: str, value: Optional[int | float | str] = None) -> None :
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        if not isinstance(value, (int, float, str)) and value is not None:
            raise TypeError(f"Expected `value` to be `int`, `float` or `str`, but got {type(value)}.")

        if value is None:
            return

        with self._lock:
            if key not in self.__table:
                self.__invalid[key] = value
                return

            item = self.__table[key]
            if not isinstance(value, item.type_):
                try:
                    value = item.type_(value)
                except ValueError:
                    raise TypeError(f"The value `{value}` is inconsistent with the constraint type.")

            if (isinstance(item.ranges, NumericalRange) and not (item.ranges.min <= value <= item.ranges.max)) or \
               (isinstance(item.ranges, Iterable) and value not in item.ranges):
                raise ConfigValueOutOfRange(f"The value `{value}` is inconsistent with the constraint ranges.")

            if item.type_ is float and isinstance(item.ranges, NumericalRange) and item.ranges != 0:
                value = round(value, item.ranges.round)

            self.__table[key].value = value

    def set_ranges(self, key: str, ranges: Iterable[int | float | str] | NumericalRange):
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        if not isinstance(ranges, (Iterable, NumericalRange)):
            raise TypeError(f"Expected `ranges` to be Iterable, but got {type(ranges)}.")

        with self._lock:
            if key not in self.__table:
                raise ConfigKeyDoesNotExist(f"The config key `{key}` does not exist.")

            item = self.__table[key]
            item.ranges = ranges

    def get(self, key: str) -> "Variable":
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        with self._lock:
            if key in self.__table:
                item = self.__table[key]
                value = item.default if item.value is None else item.value
                return _VARIABLE_CLASS_TABLE[item.type_](self, key, value)

            elif key in self.__invalid:
                return self.__invalid[key]

            raise ConfigKeyDoesNotExist(f"The config key `{key}` does not exist.")

    def get_default(self, key: str) -> Optional[int | float | str]:
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        with self._lock:
            if key not in self.__table:
                raise ConfigKeyDoesNotExist(f"The config key `{key}` does not exist.")

            item = self.__table[key]
            return item.default

    def get_ranges(self, key: str) -> Optional[Iterable[int | float | str] | NumericalRange]:
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        with self._lock:
            if key not in self.__table:
                raise ConfigKeyDoesNotExist(f"The config key `{key}` does not exist.")

            item = self.__table[key]
            return copy.copy(item.ranges)

    def get_type(self, key: str) -> Type[int | float | str]:
        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be `str`, but got {type(key)}.")

        with self._lock:
            if key not in self.__table:
                raise ConfigKeyDoesNotExist(f"The config key `{key}` does not exist.")

            item = self.__table[key]
            return item.type_

    def load_dict(self, data: Mapping) -> list[str]:
        if not isinstance(data, Mapping):
            raise TypeError(f"Expected `data` to be `Mapping`, but got {type(data)}.")

        lst = []

        for key, value in data.items():
            try:
                self.set(key, value)

            except Exception as _:
                lst.append(key)

        return lst

    def load_json(self, filepath: str, encoding: str = "utf-8", *, base64: bool = False) -> list[str]:
        with open(filepath, "r", encoding=encoding) as file_obj:
            content = file_obj.read()
            if base64:
                content = b64decode(content).decode(encoding)
            data = json.loads(content)

        with self._lock:
            self.__last.filepath = filepath
            self.__last.filetype = ConfigFileType.JSON
            self.__last.encoding = encoding
            self.__last.base64 = base64

        return self.load_dict(data)

    def save(self, filepath: Optional[str] = None, filetype: Optional[str] = None,
             encoding: Optional[str] = None, *, base64: Optional[bool] = None) -> None:
        if not isinstance(filepath, str) and filepath is not None:
            raise TypeError(f"Expected `filepath` to be `str` or None, but got {type(filepath)}.")

        if not isinstance(filetype, str) and filetype is not None:
            raise TypeError(f"Expected `filetype` to be `str` or None, but got {type(filetype)}.")

        if not isinstance(encoding, str) and encoding is not None:
            raise TypeError(f"Expected `encoding` to be `str` or None, but got {type(encoding)}.")

        if not isinstance(base64, bool) and base64 is not None:
            raise TypeError(f"Expected `base64` to be `bool` or None, but got {type(base64)}.")

        filepath = filepath or self.last_filepath
        filetype = filetype or self.last_filetype
        encoding = encoding or self.last_encoding
        base64 = self.last_base64 if base64 is None else base64

        if filepath is None or filetype is None:
            raise ConfigFileTypeNotSpecified(f"The config file type is not specified.")

        match filetype:
            case ConfigFileType.JSON:
                self.save_json(filepath, encoding, base64=base64)

            case _:
                raise ConfigFileTypeNotSupported(f"The config file type `{filetype}` is not supported.")

    def save_json(self, filepath: str, encoding: str = "utf-8", *, base64: bool = False) -> None:
        with self._lock:
            data = {key: item.value for key, item in self.__table.items()}

        if not base64:
            content = json.dumps(data, ensure_ascii=False, sort_keys=False, indent=4)

            with open(filepath, "w", encoding=encoding) as fobj:
                fobj.write(content)
            return

        content = json.dumps(data, ensure_ascii=False, sort_keys=False)
        content = b64encode(content.encode(encoding))
        with open(filepath, "wb") as fobj:
            fobj.write(content)


def setting(
        type_: Type[int | float | str],
        default: Optional[int | float | str] = None,
        ranges: Optional[Iterable[int | float | str]] = None
        ) -> Variable:
    # ! ConfigItem 会在 Configuration 中变换为 Variable 对象,
    # ! 且你不应该在 Configuration 之外使用 setting 函数,
    # ! 因此这里将函数返回类型标注为 Variable 而不是 ConfigItem 以欺骗类型检查.
    return ConfigItem(type_, default, ranges)


_VARIABLE_CLASS_TABLE = {
    int: IntVariable,
    float: FloatVariable,
    str: StrVariable
}



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

__all__ = [
    "EVENT_HUB_EMIT",
    "EVENT_HUB_SUBSCRIBE",
    "EVENT_HUB_UNSUBSCRIBE",
    "EVENT_HUB_EXCEPTION",
    "EVENT_HUB_EXEC",
    "EventHubBaseException",
    "EventNameDoesNotExist",
    "EventItem",
    "EventHubDefaultAction",
    "EventHub"
]

# std
import threading

from typing import Any, Callable, Iterable, Mapping, Optional
from dataclasses import dataclass

# site
from typex import Static, MultitonAtomic


EVENT_HUB_EMIT = "event_hub_emit"
EVENT_HUB_SUBSCRIBE = "event_hub_subscribe"
EVENT_HUB_UNSUBSCRIBE = "event_hub_unsubscribe"
EVENT_HUB_EXCEPTION = "event_hub_exception"
EVENT_HUB_EXEC = "event_hub_exec"


class EventHubBaseException (Exception): ...
class EventNameDoesNotExist (EventHubBaseException): ...


@dataclass(frozen=True)
class EventItem (object):
    iid: int
    event_name: str
    callback: Callable
    args: Iterable[Any]
    kwargs: Mapping[str, Any]
    is_async: bool


class EventHubDefaultAction (Static):
    @staticmethod
    def callback(action: str, detail: Optional[EventItem | Exception] = None) -> None:
        ...

    @staticmethod
    def exec_async_task(event_item: EventItem) -> None:
        threading.Thread(None, event_item.callback, f"{event_item.event_name} # {event_item.iid}",
                         event_item.args, event_item.kwargs).start()

    @staticmethod
    def exec_sync_task(event_item: EventItem) -> None:
        event_item.callback(*event_item.args, **event_item.kwargs)


class EventHub (object):
    def __init__(self, event_list: Optional[Iterable[str]] = None) -> None:
        if not isinstance(event_list, Iterable):
            raise TypeError(f"Expected `event_list` to be Iterable, but got {type(event_list)}.")

        self._lock = threading.RLock()
        self.__atomic = MultitonAtomic(instance_name="numlinka - ezudesign - EventHub")
        self.__event_list = []
        self.__subscribe: dict[int, EventItem] = {}
        self.__wait: dict[str, list[threading.Event]] = {}

        for event_name in event_list or []:
            self.add_event(event_name)

    def callback(self, action: str, detail: Optional[EventItem | Exception] = None) -> None:
        EventHubDefaultAction.callback(action, detail)

    def add_event(self, event_name: str) -> None:
        if not isinstance(event_name, str):
            raise TypeError(f"Expected `event_name` to be str, but got {type(event_name)}.")

        with self._lock:
            if event_name not in self.__event_list:
                self.__event_list.append(event_name)

    def wait(self, event_name: str, timeout: Optional[int | float] = None) -> bool:
        if not isinstance(event_name, str):
            raise TypeError(f"Expected `event_name` to be str, but got {type(event_name)}.")

        if not isinstance(timeout, (int, float)) and timeout is not None:
            raise TypeError(f"Expected `timeout` to be int or float, but got {type(timeout)}.")

        with self._lock:
            if event_name not in self.__event_list:
                raise EventNameDoesNotExist(f"Event name `{event_name}` does not exist.")

            if event_name not in self.__wait:
                self.__wait[event_name] = []

            unit = threading.Event()
            self.__wait[event_name].append(unit)

        return unit.wait(timeout)

    def subscribe(self, event_name: str, callback: Callable, args: Optional[Iterable[Any]] = None,
                  kwargs: Optional[Mapping[str, Any]] = None, async_: bool = False) -> int:
        if not isinstance(event_name, str):
            raise TypeError(f"Expected `event_name` to be str, but got {type(event_name)}.")

        if not callable(callback):
            raise TypeError(f"Expected `callback` to be callable, but got {type(callback)}.")

        if not isinstance(args, Iterable) and args is not None:
            raise TypeError(f"Expected `args` to be Iterable, but got {type(args)}.")

        if not isinstance(kwargs, Mapping) and kwargs is not None:
            raise TypeError(f"Expected `kwargs` to be Mapping, but got {type(kwargs)}.")

        with self._lock:
            if event_name not in self.__event_list:
                raise EventNameDoesNotExist(f"Event name `{event_name}` does not exist.")

            iid = self.__atomic.value
            item = EventItem(iid, event_name, callback, args or [], kwargs or {}, async_)
            self.__subscribe[iid] = item

        self.callback(EVENT_HUB_SUBSCRIBE, item)
        return iid

    def unsubscribe(self, anymark: int | Callable) -> bool:
        if isinstance(anymark, int):
            with self._lock:
                if anymark in self.__subscribe:
                    item = self.__subscribe[anymark]
                    del self.__subscribe[anymark]

                else:
                    return False

            self.callback(EVENT_HUB_UNSUBSCRIBE, item)

        elif callable(anymark):
            with self._lock:
                for iid, item in self.__subscribe.items():
                    if item.callback == anymark:
                        item = self.__subscribe[iid]
                        del self.__subscribe[iid]
                        break

                else:
                    return False

            self.callback(EVENT_HUB_UNSUBSCRIBE, item)

        else:
            raise TypeError(f"Expected `anymark` to be int or callable, but got {type(anymark)}.")

    def _exec_async_task(self, event_item: EventItem) -> None:
        self.callback(EVENT_HUB_EXEC, event_item)
        EventHubDefaultAction.exec_async_task(event_item)

    def _exec_sync_task(self, event_item: EventItem) -> None:
        self.callback(EVENT_HUB_EXEC, event_item)
        EventHubDefaultAction.exec_sync_task(event_item)

    def emit(self, event_name: str) -> None:
        if not isinstance(event_name, str):
            raise TypeError(f"Expected `event_name` to be str, but got {type(event_name)}.")

        self.callback(EVENT_HUB_EMIT, event_name)
        list_wait = []
        list_async: list[EventItem] = []
        list_sync: list[EventItem] = []

        with self._lock:
            if event_name not in self.__event_list:
                raise EventNameDoesNotExist(f"Event name `{event_name}` does not exist.")

            if event_name in self.__wait:
                list_wait = self.__wait[event_name]
                self.__wait[event_name] = []

            for _, item in self.__subscribe.items():
                if item.event_name == event_name:
                    if item.is_async:
                        list_async.append(item)
                    else:
                        list_sync.append(item)

        for wait_event in list_wait:
            wait_event.set()

        try:
            for item in list_async:
                self._exec_async_task(item)

            for item in list_sync:
                self._exec_sync_task(item)

        except Exception as e:
            # ! 调用错误不应由触发事件的函数承担
            self.callback(EVENT_HUB_EXCEPTION, e)

    listen = subscribe
    fire = emit



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

__all__ = ["ExecItem", "try_exec", "exec_item"]

# std
import inspect
from typing import Callable, Iterable, Mapping, Optional, Any
from dataclasses import dataclass, field

@dataclass(frozen=True)
class ExecItem (object):
    callback: Callable | str
    args: Iterable[Any] = field(default_factory=tuple)
    kwargs: Mapping[str, Any] = field(default_factory=dict)


def try_exec(exec_try: ExecItem, exec_except: Optional[ExecItem] = None) -> Any:
    if not isinstance(exec_try, (ExecItem, str)):
        raise TypeError(f"Expected `exec_try` to be ExecItem, but got {type(exec_try)}.")

    if not isinstance(exec_except, ExecItem) and exec_except is not None:
        raise TypeError(f"Expected `exec_except` to be ExecItem, but got {type(exec_except)}.")

    try:
        if isinstance(exec_try.callback, ExecItem):
            try_callable = exec_try.callback

        if isinstance(exec_try.callback, str):
            attr_names = exec_try.callback.split(".")
            frame = inspect.currentframe().f_back
            for index, attr_name in enumerate(attr_names):
                if index == 0:
                    if attr_name in frame.f_locals.keys():
                        obj = frame.f_locals[attr_name]
                    elif attr_name in frame.f_globals.keys():
                        obj = frame.f_globals[attr_name]
                    else:
                        raise NameError
                    continue
                obj = getattr(obj, attr_name)

            else:
                try_callable = obj

        return try_callable(*exec_try.args, **exec_try.kwargs)

    except Exception as e:
        if exec_except is not None:
            return exec_except.callback(e, *exec_except.args, **exec_except.kwargs)

        return e


def exec_item(callback: Callable, *args: Any, **kwargs: Any) -> ExecItem:
    return ExecItem(callback, args, kwargs)


# try_exec(exec_item("self"))



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

# std
import threading

from typing import Callable, Any, Iterable, Mapping, Optional, NoReturn
from dataclasses import dataclass, field

# site
from typex import MultitonAtomic


TASKPOOL_NEW_TASK = "task_pool_new_task"
TASKPOOL_TASK_START = "task_pool_task_start"
TASKPOOL_TASK_DONE = "task_pool_task_done"
TASKPOOL_TASK_EXCEPT = "task_pool_task_except"


@dataclass
class TaskUnit (object):
    iid: int
    task: Callable
    args: Iterable
    kwargs: Mapping
    need_return: bool
    event: threading.Event = field(default_factory=threading.Event)
    exception: Optional[Exception] = None
    result: Optional[Any] = None


@dataclass
class ThreadUnit (object):
    tpid: int
    thread: threading.Thread
    task: Optional[TaskUnit] = None
    work: threading.Event = field(default_factory=threading.Event)
    done: threading.Event = field(default_factory=threading.Event)


class TaskPool (object):
    def __init__ (self, size: int = 16) -> None:
        self._lock = threading.RLock()
        self.__queue_task: list[TaskUnit] = []
        self.__queue_running: list[TaskUnit] = []
        self.__queue_done: list[TaskUnit] = []
        self.__pool: list[ThreadUnit] = []
        self.__size: int = size

        self.__atomic = MultitonAtomic(instance_name="numlinka - ezudesign - TaskPool/TaskID")
        self.__atomic_tpid = MultitonAtomic(instance_name="numlinka - ezudesign - TaskPool/ThreadID")

    def callback(self, action: str, detail: Optional[TaskUnit | Exception] = None) -> None:
        ...

    def new_task(self, task: Callable, args: Optional[Iterable[Any]] = None,
                 kwargs: Optional[Mapping[str, Any]] = None, need_return: bool = False) -> int:
        if not callable(task):
            raise TypeError(f"Expected `task` to be callable, but got {type(task)}.")

        if not isinstance(args, Iterable) and args is not None:
            raise TypeError(f"Expected `args` to be Iterable, but got {type(args)}.")

        if not isinstance(kwargs, Mapping) and kwargs is not None:
            raise TypeError(f"Expected `kwargs` to be Mapping, but got {type(kwargs)}.")

        iid = self.__atomic.value

        unit = TaskUnit(
            iid=iid,
            task=task,
            args=args or (),
            kwargs=kwargs or {},
            need_return=need_return,
        )

        with self._lock:
            self.__queue_task.append(unit)

        self.callback(TASKPOOL_NEW_TASK, unit)
        self.__task_actvicate()
        return iid

    def join(self, iid: int, timeout: Optional[int | float] = None) -> bool:
        if not isinstance(iid, int):
            raise TypeError(f"Expected `iid` to be int, but got {type(iid)}.")

        if not isinstance(timeout, (int, float)) and timeout is not None:
            raise TypeError(f"Expected `timeout` to be int or float, but got {type(timeout)}.")

        with self._lock:
            for unit in self.__queue_task:
                if unit.iid == iid:
                    break

            else:
                for unit in self.__queue_running:
                    if unit.iid == iid:
                        break

                else:
                    return True

        unit.event.wait(timeout)

    def get_return(self, iid: int) -> Any:
        if not isinstance(iid, int):
            raise TypeError(f"Expected `iid` to be int, but got {type(iid)}.")

        with self._lock:
            for index, unit in enumerate(self.__queue_done):
                if unit.iid == iid:
                    break

            else:
                raise ValueError(f"The task with ID {iid} was not completed or was abandoned.")

            self.__queue_done.pop(index)

        return unit.result

    def __run_unit(self, unit: ThreadUnit) -> NoReturn:
        while True:
            try:
                unit.work.wait()
                unit.done.clear()

                if unit.task is None:
                    unit.work.clear()
                    unit.done.set()
                    continue

                self.callback(TASKPOOL_TASK_START, unit.task)
                result = unit.task.task(*unit.task.args, **unit.task.kwargs)
                if unit.task.need_return:
                    unit.task.result = result

            except Exception as e:
                self.callback(TASKPOOL_TASK_EXCEPT, unit.task)
                unit.task.exception = e

            else:
                self.callback(TASKPOOL_TASK_DONE, unit.task)

            finally:
                self.__task_done(unit.task)
                unit.task.event.set()
                unit.task = None
                unit.work.clear()
                unit.done.set()
                self.__task_actvicate()


    def __task_done(self, task_unit: TaskUnit) -> None:
        with self._lock:
            self.__queue_running.remove(task_unit)
            if not task_unit.need_return:
                return

            self.__queue_done.append(task_unit)

    def __task_actvicate(self) -> None:
        with self._lock:
            while True:
                if not self.__queue_task:
                    break

                for thread_unit in self.__pool:
                    if thread_unit.done.is_set():
                        break

                else:
                    if len(self.__pool) >= self.__size:
                        break
                    thread_unit = ThreadUnit(self.__atomic_tpid.value, ...)
                    self.__pool.append(thread_unit)
                    thread_unit.thread = threading.Thread(
                        None,
                        self.__run_unit,
                        f"TaskPool#{thread_unit.tpid}",
                        (thread_unit, ),
                        {},
                        daemon=True
                    )
                    thread_unit.thread.start()

                task_unit = self.__queue_task.pop(0)
                self.__queue_running.append(task_unit)
                thread_unit.done.clear()
                thread_unit.task = task_unit
                thread_unit.work.set()



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

# std
import sys
sys.path.insert(0, "src")

# site
from setuptools import setup
import ezudesign


setup(
    name = "ezudesign",
    version = ezudesign.__version__,
    description = "Some usability designs.",
    long_description = open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type = "text/markdown",
    author = "numlinka",
    author_email = "numlinka@163.com",
    url = "https://github.com/numlinka/pyezudesign",
    package_dir={"": "src"},
    packages = ["ezudesign"],
    install_requires=[
        "typex>=0.3.0"
    ],
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3",
    ],
    license = "LGPLv3",
    keywords = ["sample"]
)
